/*
** ucil_theora.c
** 
** Made by (Arne Caspari)
** Login   <arne@arne-laptop>
** 
** Started on  Tue Feb 27 07:19:39 2007 Arne Caspari
** Last update Wed Mar 21 18:37:43 2007 Arne Caspari
*/

#include "config.h"
#include <string.h>
#include <theora/theora.h>
#include <ogg/ogg.h>
#include <stdlib.h>
#include <stdarg.h>

#include "ucil_theora.h"
#include "ucil.h"
#include "unicap.h"
#if UCIL_DEBUG
#define DEBUG
#endif
#include "debug.h"
#include "../libunicap/unicap_private.h"

#define INBUF_SIZE 4096

#define ENCODEBUFFER_SIZE 4

static unsigned long long timevaltousec( struct timeval *tv )
{
   unsigned long long t = 0;
   
   t += tv->tv_sec * 1000000;
   t += tv->tv_usec;
   
   return t;
}

static int buffer_data( FILE *f, ogg_sync_state *oy )
{
   char * buffer;
   int bytes;
   buffer = ogg_sync_buffer( oy, 4096 );
   bytes = fread( buffer, 1, 4096, f );	    
   ogg_sync_wrote( oy, bytes );

   return bytes;
}

static void *ucil_theora_encode_thread( ucil_theora_video_file_object_t *vobj )
{
   yuv_buffer yuv;
   
   yuv.y_width = vobj->format.size.width;
   yuv.y_height = vobj->format.size.height;
   yuv.y_stride = vobj->format.size.width;
   yuv.uv_width = vobj->format.size.width / 2;
   yuv.uv_height = vobj->format.size.height / 2;
   yuv.uv_stride = vobj->format.size.width / 2;

   vobj->last_frame = NULL;

   while( !vobj->quit_thread )
   {
      struct _unicap_queue *entry;
      unicap_data_buffer_t *data_buffer;
      ogg_page og;
   
      entry = _get_front_queue( vobj->full_queue );
      if( !entry )
      {
	 // TODO: better synchronisation
	 usleep( 10000 );
	 continue;
      }

      data_buffer = ( unicap_data_buffer_t * ) entry->data;

      if( vobj->encode_frame_cb )
      {
	 vobj->encode_frame_cb( UNICAP_EVENT_NEW_FRAME, NULL, data_buffer, vobj->encode_frame_cb_data );
      }

      if( vobj->fill_frames )
      {
	 unsigned long long ctu;
	 unsigned long long rtu;

	 ctu = timevaltousec( &data_buffer->fill_time );
	 TRACE( "ctu: %lld\n", ctu );
	 
	 if( ctu )
	 {
	    unsigned long long etu;
	    
	    if( vobj->frame_count == 0 )
	    {
	       memcpy( &vobj->recording_start_time, &data_buffer->fill_time, sizeof( struct timeval ) );
	    }
	    else
	    {
	       if( vobj->last_frame )
	       {
		  unicap_data_buffer_t *last_data_buffer;

		  last_data_buffer = (unicap_data_buffer_t *)vobj->last_frame->data;
		  yuv.y = last_data_buffer->data;
		  yuv.u = last_data_buffer->data + ( yuv.y_stride * yuv.y_height );
		  yuv.v = yuv.u + ( yuv.uv_stride * yuv.uv_height );
		  
		  rtu = timevaltousec( &vobj->recording_start_time );
		  ctu -= rtu;

		  etu = vobj->frame_interval * (unsigned long long)vobj->frame_count + (unsigned long long)vobj->frame_interval / 2;

		  TRACE( "ctu: %lld etu: %lld\n", ctu, etu );
	       
		  while( ctu > etu )
		  {
		     TRACE( "Frame fill\n" );      
		     if( theora_encode_YUVin( &vobj->th, &yuv ) )
		     {
			TRACE( "theora_encode_YUVin FAILED!\n" );
		     }
		     theora_encode_packetout( &vobj->th, 0, &vobj->op );
		     ogg_stream_packetin( &vobj->os, &vobj->op );
		     while( ogg_stream_pageout( &vobj->os, &og ) )
		     {
			fwrite( og.header, og.header_len, 1, vobj->f );
			fwrite( og.body, og.body_len, 1, vobj->f );
		     }
		     etu += vobj->frame_interval;
		     vobj->frame_count++;
		     
		  }
		  
		  _insert_front_queue( vobj->empty_queue, vobj->last_frame );
		  vobj->last_frame = NULL;
	       }
	    }
	 }
	 else
	 {
	    vobj->fill_frames = 0;
	    if(  vobj->frame_count == 0 )
	    {
	       TRACE( "Frame time information not available; can not fill frames\n" );
	    }
	 } 
      }

      if( !vobj->fill_frames )
      {
	 if( vobj->last_frame )
	 {
	    _insert_front_queue( vobj->empty_queue, vobj->last_frame );
	 }
	 vobj->last_frame = NULL;
      }
      
      vobj->frame_count++;

      yuv.y = data_buffer->data;
      yuv.u = data_buffer->data + ( yuv.y_stride * yuv.y_height );
      yuv.v = yuv.u + ( yuv.uv_stride * yuv.uv_height );
      
      if( theora_encode_YUVin( &vobj->th, &yuv ) )
      {
	 TRACE( "theora_encode_YUVin FAILED!\n" );
      }
      memcpy( &vobj->last_frame_time, &data_buffer->fill_time, sizeof( struct timeval ) );

      vobj->last_frame = entry;
      
      theora_encode_packetout( &vobj->th, 0, &vobj->op );
      ogg_stream_packetin( &vobj->os, &vobj->op );
      
      while( ogg_stream_pageout( &vobj->os, &og ) )
      {
	 fwrite( og.header, og.header_len, 1, vobj->f );
	 fwrite( og.body, og.body_len, 1, vobj->f );
      }
   }

   if( vobj->last_frame )
   {
      // encode again to set eos
      unicap_data_buffer_t *last_data_buffer;
      ogg_page og;

      last_data_buffer = (unicap_data_buffer_t *)vobj->last_frame->data;
      yuv.y = last_data_buffer->data;
      yuv.u = last_data_buffer->data + ( yuv.y_stride * yuv.y_height );
      yuv.v = yuv.u + ( yuv.uv_stride * yuv.uv_height );
      if( theora_encode_YUVin( &vobj->th, &yuv ) )
      {
	 TRACE( "theora_encode_YUVin FAILED!\n" );
      }
      theora_encode_packetout( &vobj->th, 1, &vobj->op );
      ogg_stream_packetin( &vobj->os, &vobj->op );
      while( ogg_stream_pageout( &vobj->os, &og ) )
      {
	 fwrite( og.header, og.header_len, 1, vobj->f );
	 fwrite( og.body, og.body_len, 1, vobj->f );
      }
      _insert_front_queue( vobj->empty_queue, vobj->last_frame );
      vobj->last_frame = NULL;
   }
}

static void copy_yuv( unsigned char *dst, yuv_buffer *yuv, theora_info *ti )
{
   int y;
   unsigned char *yoff;
   unsigned char *uvoff;
   unsigned char *dstoff;
   int crop_offset;
   
   dstoff = dst;
   crop_offset = ti->offset_x + yuv->y_stride * ti->offset_y;
   yoff = yuv->y + crop_offset;
   
   for( y = 0; y < yuv->y_height; y++ )
   {
      memcpy( dstoff, yoff, yuv->y_width );
      dstoff += yuv->y_width;
      yoff += yuv->y_stride;
   }

   crop_offset = ( ti->offset_x / 2 ) + ( yuv->uv_stride ) * ( ti->offset_y / 2 );
   uvoff = yuv->u + crop_offset;

   for( y = 0; y < yuv->uv_height; y++ )
   {
      memcpy( dstoff, uvoff, yuv->uv_width );
      dstoff += yuv->uv_width;
      uvoff += yuv->uv_stride;
   }
   
   uvoff = yuv->v;
   
   for( y = 0; y < yuv->uv_height; y++ )
   {
      memcpy( dstoff, uvoff, yuv->uv_width );
      dstoff += yuv->uv_width;
      uvoff += yuv->uv_stride;
   }
}


static void *ucil_theora_worker_thread( ucil_theora_input_file_object_t *vobj )
{
   unicap_data_buffer_t new_frame_buffer;

   struct timeval ltime;
   int eos = 0;

   unicap_copy_format( &new_frame_buffer.format, &vobj->format );
   new_frame_buffer.type = UNICAP_BUFFER_TYPE_SYSTEM;
   new_frame_buffer.buffer_size = new_frame_buffer.format.buffer_size;
   new_frame_buffer.data = malloc( new_frame_buffer.format.buffer_size );

   gettimeofday( &ltime, NULL );
   
   while( !vobj->quit_capture_thread )
   {
      struct timespec abs_timeout;
      struct timeval  ctime;
      struct _unicap_queue *entry;
      ogg_page og;
      ogg_packet op;
      char * buffer;
      size_t bytes;

      int buffer_ready = 0;
      


      if( !eos && ( ogg_stream_packetout( &vobj->os, &op ) > 0 ) )
      {
	 yuv_buffer yuv;

	 theora_decode_packetin( &vobj->th, &op );
	 theora_decode_YUVout( &vobj->th, &yuv );
	 copy_yuv( new_frame_buffer.data, &yuv, &vobj->ti );

	 buffer_ready = 1;
      } 
      else if( !eos )
      {
	 bytes = buffer_data( vobj->f, &vobj->oy );      
	 if( !bytes )
	 {
	    TRACE( "End of stream\n" );
	    eos = 1;
	    
	 }
	 
	 while( ogg_sync_pageout( &vobj->oy, &og ) > 0 )
	 {
	    ogg_stream_pagein( &vobj->os, &og );
	 }
	 continue;
      }
      else
      {
	 buffer_ready = 1;
      }

      gettimeofday( &ctime, NULL );
      abs_timeout.tv_sec = ctime.tv_sec + 1;
      abs_timeout.tv_nsec = ctime.tv_usec * 1000;      
      if( sem_timedwait( &vobj->sema, &abs_timeout ) )
      {
	 TRACE( "SEM_WAIT FAILED\n" );
	 continue;
      }

      if( buffer_ready && vobj->event_callback )
      {
	 vobj->event_callback( vobj->event_unicap_handle, UNICAP_EVENT_NEW_FRAME, &new_frame_buffer );
	 TRACE( "New frame\n" );
      }
      
      entry = _get_front_queue( vobj->in_queue );
      if( entry )
      {
	 unicap_data_buffer_t *data_buffer = ( unicap_data_buffer_t * ) entry->data;

	 unicap_copy_format( &data_buffer->format, &vobj->format );
	 memcpy( data_buffer->data, new_frame_buffer.data, vobj->format.buffer_size );
	 
	 _insert_back_queue( vobj->out_queue, entry );
      }

      sem_post( &vobj->sema );
      
      if( buffer_ready )
      {
	 gettimeofday( &ctime, NULL );
	 if( ctime.tv_usec < ltime.tv_usec )
	 {
	    ctime.tv_usec += 1000000;
	    ctime.tv_sec -= 1;
	 }
	 
	 ctime.tv_usec -= ltime.tv_usec;
	 ctime.tv_sec -= ltime.tv_sec;
	 
	 if( ( ctime.tv_sec == 0 ) &&
	     ( ctime.tv_usec < vobj->frame_intervall ) )
	 {
	    usleep( vobj->frame_intervall - ctime.tv_usec );
	 }
      
	 gettimeofday( &ltime, NULL );
      }
   }

   free( new_frame_buffer.data );
}

static unicap_status_t theoracpi_reenumerate_formats( ucil_theora_input_file_object_t vobj, int *count )
{
   *count = 1;

   return STATUS_SUCCESS;
}

static unicap_status_t theoracpi_enumerate_formats( ucil_theora_input_file_object_t *vobj, unicap_format_t *format, int index )
{
   unicap_status_t status = STATUS_NO_MATCH;
   if( index == 0 )
   {
      unicap_copy_format( format, &vobj->format );
      status = STATUS_SUCCESS;
   }
   
   return status;
}

static unicap_status_t theoracpi_set_format( ucil_theora_input_file_object_t *vobj, unicap_format_t *format )
{
   unicap_status_t status = STATUS_SUCCESS;
   if( ( format->size.width != vobj->format.size.width ) || 
       ( format->size.height != vobj->format.size.height ) ||
       ( format->bpp != vobj->format.bpp ) )
   {
      char buffer[1024];
      int size = 1024;

      unicap_describe_format( format, buffer, &size );
      TRACE( "Could not set format: %s\n", buffer );
      size = 1024;
      unicap_describe_format( &vobj->format, buffer, &size );
      TRACE( "Stored: %s\n" );
      status = STATUS_FAILURE;
   }   
   
   return status;
}

static unicap_status_t theoracpi_get_format( ucil_theora_input_file_object_t *vobj, unicap_format_t *format )
{
   unicap_copy_format( format, &vobj->format );
   return STATUS_SUCCESS;
}

static unicap_status_t theoracpi_reenumerate_properties( ucil_theora_input_file_object_t *vobj, int *count )
{
   *count = 0;
   return STATUS_SUCCESS;
}

static unicap_status_t theoracpi_enumerate_properties( ucil_theora_input_file_object_t *vobj, unicap_property_t *property, int index )
{
   return STATUS_NO_MATCH;
}

static unicap_status_t theoracpi_set_property( ucil_theora_input_file_object_t *vobj, unicap_property_t *property )
{
   return STATUS_FAILURE;
}

static unicap_status_t theoracpi_get_property( ucil_theora_input_file_object_t *vobj, unicap_property_t *property )
{
   return STATUS_FAILURE;
}

static unicap_status_t theoracpi_capture_start( ucil_theora_input_file_object_t *vobj )
{
   unicap_status_t status = STATUS_SUCCESS;
   
   if( pthread_create( &vobj->worker_thread, NULL, (void*(*)(void*))ucil_theora_worker_thread, vobj ) )
   {
      TRACE( "Failed to create worker thread!\n" );
      return STATUS_FAILURE;
   }

   return status;
}

static unicap_status_t theoracpi_capture_stop( ucil_theora_input_file_object_t *vobj )
{
   unicap_status_t status = STATUS_SUCCESS;
   int res;

   vobj->quit_capture_thread = 1;
   res = pthread_join( vobj->worker_thread, NULL );

   return ( res == 0 ) ? STATUS_SUCCESS : STATUS_FAILURE;
}

static unicap_status_t theoracpi_queue_buffer( ucil_theora_input_file_object_t *vobj, unicap_data_buffer_t *buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   struct _unicap_queue *entry;
   
   entry = malloc( sizeof( struct _unicap_queue ) );
   entry->data = buffer;
   _insert_back_queue( vobj->in_queue, entry );

   return status;
}

static unicap_status_t theoracpi_dequeue_buffer( ucil_theora_input_file_object_t *vobj, unicap_data_buffer_t **buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   return status;
}

static unicap_status_t theoracpi_wait_buffer( ucil_theora_input_file_object_t *vobj, unicap_data_buffer_t **buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   return status;
}

static unicap_status_t theoracpi_poll_buffer( ucil_theora_input_file_object_t *vobj, int *count )
{
   *count = 1;
   return STATUS_SUCCESS;
}

static unicap_status_t theoracpi_set_event_notify( ucil_theora_input_file_object_t *vobj, unicap_event_callback_t func, unicap_handle_t handle )
{
   vobj->event_callback = func;
   vobj->event_unicap_handle = handle;

   return STATUS_SUCCESS;
}

static void encode_parse_args( ucil_theora_video_file_object_t *vobj, va_list ap )
{
   char *arg;

   while( ( arg = va_arg( ap, char * ) ) )
   {
      if( !strcmp( arg, "fill_frames" ) )
      {
	 vobj->fill_frames = va_arg( ap, int );
	 TRACE( "fill_frames: %d\n", vobj->fill_frames );
      }
      else if( !strcmp( arg, "fps" ) )
      {
	 unsigned int num;
	 double fps;
	 
	 fps = va_arg( ap, double );
	 num = fps * 1000000;
	 
	 vobj->ti.fps_numerator = num;
	 vobj->ti.fps_denominator = 1000000;	 
      }
      else if( !strcmp( arg, "quality" ) )
      {
	 vobj->ti.quality = va_arg( ap, int );
      }
      else if( !strcmp( arg, "bitrate" ) )
      {
	 int bitrate;
	 
	 bitrate = va_arg( ap, int );
	 
	 vobj->ti.target_bitrate = bitrate;
	 vobj->ti.keyframe_data_target_bitrate = bitrate * 2;
      }
      else if( !strcmp( arg, "encode_frame_cb" ) )
      {
	 unicap_new_frame_callback_t cb;
	 
	 vobj->encode_frame_cb = va_arg( ap, unicap_new_frame_callback_t );
	 vobj->encode_frame_cb_data = va_arg( ap, void* );
      }
      else
      {
	 TRACE( "Unknown arg: %s\n", arg );
	 break;
      }
   }
}

ucil_theora_video_file_object_t *ucil_theora_create_video_file( const char *path, unicap_format_t *format, va_list ap)
{
   ucil_theora_video_file_object_t *vobj;   
   ogg_packet op;
   ogg_page og;
   theora_comment tc;
   int i;
   

   vobj = malloc( sizeof( ucil_theora_video_file_object_t ) );
   memset( vobj, 0x0, sizeof( ucil_theora_video_file_object_t ) );

   vobj->f = fopen( path, "w" );
   if( !vobj->f )
   {
      free( vobj );
      return NULL;
   }

   if( ogg_stream_init( &vobj->os, 1 ) != 0 )
   {
      TRACE( "ogg_stream_init failed!\n" );
      free( vobj );
      return NULL;
   }

   theora_info_init( &vobj->ti );

   vobj->ti.width = format->size.width;
   vobj->ti.height = format->size.height;
   vobj->ti.frame_width = format->size.width;
   vobj->ti.frame_height = format->size.height;
   vobj->ti.offset_x = 0;
   vobj->ti.offset_y = 0;
   vobj->ti.fps_numerator = 30000000;
   vobj->ti.fps_denominator = 1000000;
   vobj->ti.aspect_numerator = 1;//format->size.width;
   vobj->ti.aspect_denominator = 1;//format->size.height;
   vobj->ti.colorspace = OC_CS_UNSPECIFIED;
   vobj->ti.pixelformat = OC_PF_420;
   vobj->ti.target_bitrate = 800000;
   vobj->ti.quality = 16;
   vobj->ti.quick_p = 1;
   vobj->ti.dropframes_p = 0;
   vobj->ti.keyframe_auto_p = 1;
   vobj->ti.keyframe_frequency=64;
   vobj->ti.keyframe_frequency_force=64;
   vobj->ti.keyframe_data_target_bitrate=800000*2;
   vobj->ti.keyframe_auto_threshold=80;
   vobj->ti.keyframe_mindistance=8;
   vobj->ti.noise_sensitivity=1;

   encode_parse_args( vobj, ap );
   
   theora_encode_init( &vobj->th, &vobj->ti );   

   vobj->full_queue = malloc( sizeof( struct _unicap_queue ) );
   _init_queue( vobj->full_queue );
   vobj->empty_queue = malloc( sizeof( struct _unicap_queue ) );
   _init_queue( vobj->empty_queue );

   for( i = 0; i < ENCODEBUFFER_SIZE; i++ )
   {
      struct _unicap_queue *entry;
      unicap_data_buffer_t *data_buffer;
      
      entry = malloc( sizeof( struct _unicap_queue ) );
      data_buffer = malloc( sizeof( unicap_data_buffer_t ) );
      unicap_copy_format( &data_buffer->format, format );
      data_buffer->format.fourcc = UCIL_FOURCC( 'I', '4', '2', '0' );
      data_buffer->format.bpp = 16;
      data_buffer->buffer_size = format->size.width * format->size.height * 2;
      unicap_copy_format( &vobj->format, format );
      
      data_buffer->data = malloc( format->size.width * format->size.height * 1.5 );
      data_buffer->buffer_size = format->size.width * format->size.height * 1.5;

      entry->data = data_buffer;
      
      _insert_front_queue( vobj->empty_queue, entry );
   }
   
   theora_encode_header( &vobj->th, &op );
   ogg_stream_packetin( &vobj->os, &op );
   if( ogg_stream_pageout( &vobj->os, &og ) != 1 )
   {
      TRACE( "Internal Ogg error\n" );
      return NULL;
   }
   fwrite( og.header, 1, og.header_len, vobj->f );
   fwrite( og.body, 1, og.body_len, vobj->f );
   theora_comment_init(&tc);
   theora_encode_comment(&tc,&op);
   ogg_stream_packetin(&vobj->os,&op);
   theora_encode_tables(&vobj->th,&op);
   ogg_stream_packetin(&vobj->os,&op);

   while( ogg_stream_flush( &vobj->os, &og ) > 0 )
   {
      fwrite( og.header, 1, og.header_len, vobj->f );
      fwrite( og.body, 1, og.body_len, vobj->f );
   }

   vobj->quit_thread = 0;
   if( pthread_create( &vobj->encoder_thread, NULL, (void*(*)(void*))ucil_theora_encode_thread, vobj ) )
   {
      TRACE( "Failed to create encoder thread!\n " );
   }

   vobj->frame_interval = 30000;
   
   return vobj;
}

unicap_status_t ucil_theora_encode_frame( ucil_theora_video_file_object_t *vobj, unicap_data_buffer_t *buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   struct _unicap_queue *entry;
   
   entry = _get_front_queue( vobj->empty_queue );
   if( !entry )
   {
      status = STATUS_FAILURE;
   }
   else
   {
      unicap_data_buffer_t *dst_buffer = (unicap_data_buffer_t*)entry->data;
      
      ucil_convert_buffer( dst_buffer, buffer );
      memcpy( &dst_buffer->fill_time, &buffer->fill_time, sizeof( struct timeval ) );
      _insert_back_queue( vobj->full_queue, entry );
   }

   return status;
}

unicap_status_t ucil_theora_close_video_file( ucil_theora_video_file_object_t *vobj )
{
   struct _unicap_queue *entry;
   
   vobj->quit_thread = 1;
   pthread_join( vobj->encoder_thread, NULL );

   for( entry = _get_front_queue( vobj->full_queue ); entry; entry = _get_front_queue( vobj->full_queue ) )
   {
      unicap_data_buffer_t *data_buffer = (unicap_data_buffer_t*) entry->data;
      free( data_buffer->data );
      free( data_buffer );
      free( entry );
   }
   free( vobj->full_queue );
   for( entry = _get_front_queue( vobj->empty_queue ); entry; entry = _get_front_queue( vobj->empty_queue ) )
   {
      unicap_data_buffer_t *data_buffer = (unicap_data_buffer_t*) entry->data;
      free( data_buffer->data );
      free( data_buffer );
      free( entry );
   }
   free( vobj->empty_queue );

   fclose( vobj->f );

   theora_info_clear( &vobj->ti );

   free( vobj );
}


unicap_status_t ucil_theora_open_video_file( unicap_handle_t *unicap_handle, char *filename )
{
   struct _unicap_handle *handle;
   ogg_page og;
   ogg_packet op;
   theora_comment tc;
   int ret;
   int pkt = 0;
   int i;
   ucil_theora_input_file_object_t *vobj;
   
   
   handle = malloc( sizeof( struct _unicap_handle ) );
   if( !handle )
   {
      return STATUS_FAILURE;
   }
   // initialize data structure
   memset( handle, 0, sizeof( struct _unicap_handle ) );
   handle->cb_info = malloc( sizeof( struct _unicap_callback_info ) * UNICAP_EVENT_LAST );
   for( i = 0; i < UNICAP_EVENT_LAST; i++ )
   {
      handle->cb_info[ i ].func = NULL;
      handle->cb_info[ i ].user_ptr = NULL;
   }

   handle->sem_key = -1;
   handle->sem_id = -1;
   
   handle->ref_count = malloc( sizeof( int ) );
   *(handle->ref_count) = 1;

   vobj = malloc( sizeof( ucil_theora_input_file_object_t ) );
   if( !vobj )
   {
      free( handle );
      return STATUS_FAILURE;
   }   

   memset( vobj, 0x0, sizeof( ucil_theora_input_file_object_t ) );


   vobj->f = fopen( filename, "r" );
   if( !vobj->f )
   {
      free( vobj );
      free( handle );
      return STATUS_FAILURE;
   }

   ogg_sync_init( &vobj->oy );
   theora_comment_init( &tc );
   theora_info_init( &vobj->ti );

   
   for( pkt = 0; pkt < 3;  )
   {   
      
      if( ogg_sync_pageout( &vobj->oy, &og ) > 0 )
      {
	 if( pkt == 0 )
	 {
	    if( !ogg_page_bos( &og ) )
	    {
	       TRACE( "Not a valid ogg stream\n" );
	       // TODO: fix page leak
	       return STATUS_FAILURE;
	    }
	    ogg_stream_init( &vobj->os, ogg_page_serialno( &og ) );
	 }
	 if( ogg_stream_pagein( &vobj->os, &og ) < 0 )
	 {
	    TRACE( "Failed to decode first page\n" );
	    fclose( vobj->f );
	    return STATUS_FAILURE;
	 }

	 while( ( pkt < 3 ) && ogg_stream_packetout( &vobj->os, &op ) > 0 )
	 {
	    if( theora_decode_header( &vobj->ti, &tc, &op ) != 0 )
	    {
	       TRACE( "Failed to decode header packet\n" );
	       fclose( vobj->f );
	       return STATUS_FAILURE;
	    }
	    pkt ++;
	 }
      }
      else
      {
	 int bytes;
	 bytes = buffer_data( vobj->f, &vobj->oy );
	 
	 if( bytes == 0 )
	 {
	    TRACE( "End of file while parsing headers\n" );
	    fclose( vobj->f );
	    return STATUS_FAILURE;
	 }  
      }
   }
   
   theora_comment_clear( &tc );

   unicap_void_format( &vobj->format );   
   strcpy( vobj->format.identifier, "UCIL default format" );
   vobj->format.size.width = vobj->ti.frame_width;
   vobj->format.size.height = vobj->ti.frame_height;
   switch( vobj->ti.pixelformat )
   {
      case OC_PF_420: 
	 vobj->format.fourcc = UCIL_FOURCC( 'I', '4', '2', '0' );
	 vobj->format.bpp = 12;
	 break;
	 
      case OC_PF_422:
	 vobj->format.fourcc = UCIL_FOURCC( 'U', 'Y', 'V', 'Y' );
	 vobj->format.bpp = 16;
	 break;
	 
      case OC_PF_444:
	 vobj->format.fourcc = UCIL_FOURCC( 'I', '4', '4', '4' );
	 vobj->format.bpp = 24;
	 break;
   }

   vobj->format.buffer_size = vobj->format.size.width * vobj->format.size.height * vobj->format.bpp / 8;
   vobj->format.buffer_type = UNICAP_BUFFER_TYPE_SYSTEM;	    

   theora_decode_init( &vobj->th, &vobj->ti );

   handle->cpi.cpi_register = NULL;
   handle->cpi.cpi_version = 1<<16;
   handle->cpi.cpi_capabilities = 0x3ffff;
   handle->cpi.cpi_enumerate_devices = NULL;
   handle->cpi.cpi_open = NULL;
   handle->cpi.cpi_close = NULL;
   handle->cpi.cpi_reenumerate_formats = (cpi_reenumerate_formats_t ) theoracpi_reenumerate_formats;
   handle->cpi.cpi_enumerate_formats = ( cpi_enumerate_formats_t ) theoracpi_enumerate_formats;
   handle->cpi.cpi_set_format = ( cpi_set_format_t ) theoracpi_set_format;
   handle->cpi.cpi_get_format = ( cpi_get_format_t ) theoracpi_get_format;
   handle->cpi.cpi_reenumerate_properties = ( cpi_reenumerate_properties_t ) theoracpi_reenumerate_properties;
   handle->cpi.cpi_enumerate_properties = ( cpi_enumerate_properties_t ) theoracpi_enumerate_properties;
   handle->cpi.cpi_get_property = ( cpi_get_property_t ) theoracpi_get_property;
   handle->cpi.cpi_set_property = ( cpi_set_property_t ) theoracpi_set_property;
   handle->cpi.cpi_capture_start = ( cpi_capture_start_t ) theoracpi_capture_start;
   handle->cpi.cpi_capture_stop = ( cpi_capture_stop_t ) theoracpi_capture_stop;
   handle->cpi.cpi_queue_buffer = ( cpi_queue_buffer_t ) theoracpi_queue_buffer;
   handle->cpi.cpi_dequeue_buffer = ( cpi_dequeue_buffer_t ) theoracpi_dequeue_buffer;
   handle->cpi.cpi_wait_buffer = ( cpi_wait_buffer_t ) theoracpi_wait_buffer;
   handle->cpi.cpi_poll_buffer = ( cpi_poll_buffer_t ) theoracpi_poll_buffer;
   handle->cpi.cpi_set_event_notify = ( cpi_set_event_notify_t ) theoracpi_set_event_notify;
   handle->cpi_data = vobj;

   unicap_cpi_register_event_notification( theoracpi_set_event_notify, vobj, handle );   

   vobj->frame_intervall = 1000000 / ( vobj->ti.fps_numerator / vobj->ti.fps_denominator );

   sem_init( &vobj->sema, 0, 1 );

   *unicap_handle = handle;

   vobj->in_queue = malloc( sizeof( struct _unicap_queue ) );
   _init_queue( vobj->in_queue );
   vobj->out_queue = malloc( sizeof( struct _unicap_queue ) );
   _init_queue( vobj->out_queue );

   return STATUS_SUCCESS;
}
